/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *    
 *   Copyright (C) 2001-2004  by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or 
 *   modify it under the terms of the GNU General Public License as 
 *   published by the Free Software Foundation; either version 2 of 
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details. 
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@gmx.de
 *    Borg Enders  borg@borgsoft.de
 *
 *********************************************************************/

#ifdef USE_NETWORK

#ifndef CLASS_NETWORK_CONNECTION_HEADER
#define CLASS_NETWORK_CONNECTION_HEADER

#include "../os/os.h"
#include "types.h"

namespace Network {

  class Server;

  /**
   ** base class for a network connection
   **
   ** @author	Diether Knof
   **/
  class Connection : public OS {
    public:
      // create a new connection
      static Connection* new_(Server& server,
			      string const& address, unsigned const port);
    protected:
      // constructor
      Connection(Server& server);
    public:
      // destructor
      virtual ~Connection();

      // the server
      Server& server();

      // sends the text over the connection
      //   Important: the text must be appendet a newline
      //   Important: call of '::ui->network_traffic()' in the child classes
      virtual void send(string const& text) = 0;

    protected:
      // received
      void received(string const& text);

      // close this connection
      void close();

    public:
      // send the data of this player (t.i. p.e. the name)
      void send_self_data();

    private:
      // analyses the received text and act
      bool analyse_and_act(string const& text);

    private:
      // the corresponding server
      Server* const server_;

      // the name of the connection
      PROT_VAR_R(string, name);
      // the address of the connected computer
      PROT_VAR_R(string, address);
      // the port on the connected computer
      PROT_VAR_R(unsigned, port);
      // whether the connection is accepted
      PROT_VAR_R(bool, accepted);

    private: // unused
      Connection();
      Connection(Connection const&);
      Connection& operator=(Connection const&);
  }; // class Connection : public OS

} // namespace Network

#endif // #ifndef CLASS_NETWORK_CONNECTION_HEADER

#endif // #ifdef USE_NETWORK
