/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *    
 *   Copyright (C) 2001-2004  by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or 
 *   modify it under the terms of the GNU General Public License as 
 *   published by the Free Software Foundation; either version 2 of 
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details. 
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@gmx.de
 *    Borg Enders  borg@borgsoft.de
 *
 **********************************************************************/

#ifndef CLASS_OS_BUG_REPORT_HEADER
#define CLASS_OS_BUG_REPORT_HEADER

#include "os.h"

namespace OS_NS {

  extern string const BUG_REPORT_FILE;

  // The default BugReport
  class BugReport : public OS {
    public:
      string report(string const& message = NULL);

    public:
      BugReport();

      ~BugReport();

      // add text to the bug report
      void add(string const& text);

      // the parts of a game
      void game_open(Game const& game);
      void game_redistribute();
      void new_game();
      void game_start();
      void game_finish();

      void trick_open(Trick const& trick);
      void trick_full();
      // the card is played
      void card_played(HandCard const& card);
      // an announcement is made
      void announcement_made(Announcement const& announcement,
				     Player const& player);
      // the player has swines
      void swines_announced(Player const& player);
      // the player has hyperswines
      void hyperswines_announced(Player const& player);

      // 'player' shifts 'cardno' cards
      void poverty_shift(Player const& player, unsigned const cardno);
      // the player 'player' has denied the poverty trumps
      void poverty_take_denied(Player const& player);
      // all players have denied to take the cards
      void poverty_take_denied_by_all();
      // the player 'player' has accepted the poverty trumps
      // and has returned 'cardno' cards with 'trumpno' trumps
      void poverty_take_accepted(Player const& player,
					 unsigned const cardno,
					 unsigned const trumpno);

      // a player has announced a genscher
      void genscher(Player const& genscher, Player const& partner);

      string directory() const;
      string filename() const;
      string filename(string const directory) const;
    private:
      void stream_clear();
      void backup_clear();

    public:
      ostream* ostr;
    private:
      ostream* ostr_human;
      unsigned seed;

      // backup
      ostream* ostr_bak;
      ostream* ostr_human_bak;
      unsigned seed_bak;

    private: // not implemented
      BugReport(BugReport const&);
      BugReport& operator=(BugReport const&);
  }; // class BugReport

} // namespace OS_NS

#endif // #ifndef CLASS_OS_BUG_REPORT_HEADER
